"""Base class for normalizer."""

import pandas as pd
import sklearn.discriminant_analysis as sk_da
import sklearn.preprocessing as sk_pre


class Normalizer:
    """
    A base class for normalizing data.

    This class provides a template for normalizing data using scikit-learn's StandardScaler and normalize functions.
    """

    def __init__(self) -> None:
        self._scaler = sk_da.StandardScaler()

    def normalize(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Normalize the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The data to normalize.

        Returns
        -------
        pd.DataFrame: The normalized data.
        """
        # scale all features (columns) so that they become comparable
        d_scaled = self._scaler.fit_transform(data)
        # normalize the data so that it approximately follows a Gaussian distribution
        d_normalized = sk_pre.normalize(d_scaled)
        # return the normalized data as a DataFrame
        return pd.DataFrame(d_normalized, index=data.index)



"""Utility functions for DataFrame manipulation."""

import itertools
import operator
import typing

import pandas as pd


def merge_dataframe_by_columns(
    *,
    datas: typing.Iterable[pd.DataFrame],
) -> pd.DataFrame:
    """
    Merge multiple DataFrames by columns.

    Parameters
    ----------
    datas : typing.Iterable[pd.DataFrame]
        The DataFrames to merge.

    Returns
    -------
    pd.DataFrame
        The merged DataFrame.
    """
    output_df = pd.concat(datas, axis="columns")

    return output_df


def add_level_to_index(
    *,
    index: typing.Union[pd.MultiIndex, pd.Index],
    new_level: typing.Hashable,
    new_level_name: str,
) -> pd.MultiIndex:
    """
    Add a new level to the index of a DataFrame.

    Parameters
    ----------
    index : typing.Union[pd.MultiIndex, pd.Index]
        The index to add a level to.
    new_level : typing.Hashable
        The value to add to the new level.
    new_level_name : str
        The name of the new level.

    Returns
    -------
    pd.MultiIndex
        The new MultiIndex.
    """
    names = index.names + [new_level_name]
    if isinstance(index, pd.MultiIndex):
        index = index.to_flat_index()
    return pd.MultiIndex.from_tuples(
        [val + (new_level,) for val in index.array], names=names
    )


def reorder_index(
    *,
    index: typing.Union[pd.MultiIndex, pd.Index],
    orders: typing.List[typing.Union[str, int]],
) -> pd.MultiIndex:
    """
    Reorder the levels of a MultiIndex.

    Parameters
    ----------
    index : typing.Union[pd.MultiIndex, pd.Index]
        The index to reorder.
    orders : typing.List[typing.Union[str, int]]
        The new order of the levels.

    Returns
    -------
    pd.MultiIndex
        The reordered MultiIndex.
    """
    if isinstance(index, pd.MultiIndex):
        return index.reorder_levels(orders)
    return index


def sort_columns(
    *,
    data: pd.DataFrame,
    level: typing.Union[str, int],
    pressing: typing.List[typing.Union[str, int]],
) -> pd.DataFrame:
    """
    Sort the columns of a DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to sort.
    level : typing.Union[str, int]
        The level to sort on.
    pressing : typing.List[typing.Union[str, int]]
        The columns to press to the front.

    Returns
    -------
    pd.DataFrame
        The sorted DataFrame.
    """
    idx = data.columns
    pressed = set(pressing)
    # get correct ordering of the index
    level_ordering = pressing + list(
        # filter out the pressing columns
        map(
            operator.itemgetter(0),
            itertools.groupby(
                sorted(
                    v for v in idx.get_level_values(level).values if v not in pressed
                )
            ),
        )
    )

    if not isinstance(idx, pd.MultiIndex):
        return data.reindex(columns=level_ordering, level=level)

    # recreate Multiindex with mapping and reindex all columns
    new_level_index = data.reindex(level=level, columns=level_ordering).columns
    new_column_order_map = dict(
        zip(  # noqa: B905
            idx[idx.isin(new_level_index)],
            new_level_index,
            # strict=True, TODO: Remove this when python3.9 is suppport is dropped
        )
    )
    new_multi_index = pd.MultiIndex.from_tuples(
        [new_column_order_map.get(x, x) for x in idx],
        names=idx.names,
    )
    return data.reindex(new_multi_index, axis=1)



"""Hierarchical Clustering."""

import typing as t

import pandas as pd
import sklearn.cluster as sk_cl

# using sklearn Agglomerative Clustering
# https://scikit-learn.org/stable/modules/generated/sklearn.cluster.AgglomerativeClustering.html


class HierarchicalClustering:
    """
    A class to perform hierarchical clustering using scikit-learn's AgglomerativeClustering.

    Parameters
    ----------
    - number_clusters (int): The number of clusters to find. Default is 2.
    - linkage (str): The linkage criterion to use. One of {'ward', 'complete', 'average', 'single'}. Default is 'ward'.
    - metric (str or callable): The metric to use when calculating distance between instances in a feature array. Default is 'euclidean'.

    Methods
    -------
    - cluster(data: pd.DataFrame) -> t.List[int]: Performs hierarchical clustering on the provided data and returns the cluster labels.
    """

    def __init__(
        self,
        number_clusters: int = 2,
        linkage: t.Literal["ward", "complete", "average", "single"] = "ward",
        metric: t.Union[
            t.Literal["euclidean", "l1", "l2", "manhattan", "cosine", "precomputed"],
            t.Callable,
        ] = "euclidean",
    ) -> None:
        self._n_clusters = number_clusters
        self._linkage = linkage
        self._metric = metric

    def cluster(self, data: pd.DataFrame) -> t.List[int]:
        """
        Perform hierarchical clustering on the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The data to cluster.

        Returns
        -------
        - t.List[int]: The cluster labels.
        """
        ac = sk_cl.AgglomerativeClustering(
            n_clusters=self._n_clusters, linkage=self._linkage, metric=self._metric
        )
        return ac.fit_predict(data)



"""
CPA (Cluster Pattern Analysis) module.

The CPA stage is responsible for analyzing cluster patterns within the data. It includes functions for:
- Data preprocessing
- Statistical analysis

Modules:
- normalizer: Functions for data normalization
- clustering: Functions for clustering data
"""



"""Principal Component Analysis (PCA) Normalizer."""

import pandas as pd
import sklearn.decomposition as sk_dc

from lta.cores.cpa.normalizer.base import Normalizer


class PrincipalComponentAnalysis(Normalizer):
    """
    A class to perform normalization and dimensionality reduction using Principal Component Analysis (PCA).

    This class extends the Normalizer base class to first normalize the data and then apply PCA to reduce its dimensionality.

    Parameters
    ----------
    - output_dimensions (int): The number of principal components to keep. Default is 2.

    """

    def __init__(self, output_dimensions: int = 2) -> None:
        super().__init__()
        self._pca = sk_dc.PCA(n_components=output_dimensions)

    def normalize(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Normalize the provided data and reduce its dimensionality using PCA.

        Parameters
        ----------
        data : pd.DataFrame
            The data to normalize and reduce dimensionality.

        Returns
        -------
        pd.DataFrame: The normalized and reduced data.
        """
        normalized = super().normalize(data)
        # reduce the dimensionality of the data
        d_principal = self._pca.fit_transform(normalized)
        return pd.DataFrame(d_principal, index=data.index)
